Benchmarks
---

*Dislcaimer: remember benchmarks are generally fraught with issues and their
legitimacy depends on a multitude of factors.*

I've used basic, but seemingly nice, [`pytest-benchmark`](https://pytest-benchmark.readthedocs.io/en/stable/usage.html)
to produce the following benchmarks on a fairly modern machine.

If you'd like to run these yourself, you'll need whatever according system 
dependency, if any, required by the specific algorithm as well as the Python
lib installed. Can install all via `pip install -r benchmark-requirements.txt`. 

The resulting output from benchmarks indicate what file and lib was used, (`cramjam` vs `gzip`)
`test_gzip[urls.10K-cramjam]` indicates `cramjam` and `benchmarks/data/urls.10K` file was used during the gzip benchmark.


Special performance notes:
---
In general, if the length of the output from de/compression is known, for any variant,  
you can provide `output_len=<<some int>>` to get a 1-4x performance increase. This
is allows us to avoid double allocation. 

Furthermore, you can provide `output_len` equal to the "at most" length when 
de/compressing `bytearray` objects; as these can be resized after the de/compression 
is finished and actual size is known for basically the same performance. This is not
true for `bytes` objects, as we cannot resize after de/compression and the result
will be trialing null bytes.

For snappy used in conjunction with `bytearray`s, this is only midly helpful as
we automatically estimate the buffer size and can resize the `bytearray` after.


--- 

#### Gzip

`make bench-gzip`


Notes
---
This benchmark is parameterized over setting `output_len=True/False` as an example
of the performance benefit that can be expected from various files/sizes.


```bash 
---------------------------------------------------------------------------------------------------------------------- benchmark: 48 tests ----------------------------------------------------------------------------------------------------------------------
Name (time in us)                                                              Min                    Max                   Mean                StdDev                 Median                   IQR            Outliers         OPS            Rounds  Iterations
-----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
test_gzip[Mark.Twain-Tom.Sawyer.txt-used-output_len=False-cramjam]        498.6320 (1.0)         904.3340 (1.02)        538.3901 (1.02)        48.5933 (1.17)        524.5310 (1.01)        37.6930 (1.09)      152;105  1,857.3891 (0.98)       1505           1
test_gzip[Mark.Twain-Tom.Sawyer.txt-used-output_len=False-gzip]           574.8140 (1.15)      1,032.3880 (1.16)        616.4469 (1.17)        53.9548 (1.30)        602.6820 (1.17)        45.1167 (1.31)       122;85  1,622.1997 (0.86)       1475           1
test_gzip[Mark.Twain-Tom.Sawyer.txt-used-output_len=True-cramjam]         499.6490 (1.00)        887.7370 (1.0)         528.1259 (1.0)         41.6166 (1.0)         517.1500 (1.0)         34.4905 (1.0)        149;93  1,893.4877 (1.0)        1676           1
test_gzip[Mark.Twain-Tom.Sawyer.txt-used-output_len=True-gzip]            575.3320 (1.15)        961.8130 (1.08)        617.2423 (1.17)        48.7969 (1.17)        601.3870 (1.16)        41.9803 (1.22)       127;71  1,620.1094 (0.86)       1329           1
test_gzip[alice29.txt-used-output_len=False-cramjam]                   10,896.4080 (21.85)    12,870.5330 (14.50)    11,282.3297 (21.36)      330.3431 (7.94)     11,201.8925 (21.66)      364.0470 (10.55)        15;3     88.6342 (0.05)         86           1
test_gzip[alice29.txt-used-output_len=False-gzip]                      12,027.0880 (24.12)    14,208.4480 (16.01)    12,427.6071 (23.53)      395.2144 (9.50)     12,312.2350 (23.81)      363.9955 (10.55)        12;3     80.4660 (0.04)         80           1
test_gzip[alice29.txt-used-output_len=True-cramjam]                    10,912.5080 (21.88)    13,144.6680 (14.81)    11,313.8446 (21.42)      331.2081 (7.96)     11,236.6165 (21.73)      315.2310 (9.14)         11;3     88.3873 (0.05)         90           1
test_gzip[alice29.txt-used-output_len=True-gzip]                       11,991.7980 (24.05)    14,076.5000 (15.86)    12,457.9527 (23.59)      439.7621 (10.57)    12,263.0020 (23.71)      461.3358 (13.38)        11;5     80.2700 (0.04)         81           1
test_gzip[asyoulik.txt-used-output_len=False-cramjam]                   8,340.2480 (16.73)     9,570.2210 (10.78)     8,648.5055 (16.38)      233.7639 (5.62)      8,591.8990 (16.61)      269.5990 (7.82)         27;5    115.6269 (0.06)        118           1
test_gzip[asyoulik.txt-used-output_len=False-gzip]                      9,158.9250 (18.37)     9,988.6440 (11.25)     9,459.7682 (17.91)      186.1147 (4.47)      9,412.7690 (18.20)      276.7113 (8.02)         38;0    105.7108 (0.06)        107           1
test_gzip[asyoulik.txt-used-output_len=True-cramjam]                    8,319.9830 (16.69)    10,737.6260 (12.10)     8,666.3427 (16.41)      322.9455 (7.76)      8,599.2010 (16.63)      258.3250 (7.49)         12;3    115.3889 (0.06)        113           1
test_gzip[asyoulik.txt-used-output_len=True-gzip]                       9,188.2840 (18.43)    11,275.7290 (12.70)     9,525.4683 (18.04)      351.3017 (8.44)      9,419.5680 (18.21)      268.0395 (7.77)          8;6    104.9817 (0.06)        100           1
test_gzip[fireworks.jpeg-used-output_len=False-cramjam]                 3,934.7320 (7.89)      6,228.3300 (7.02)      4,123.2676 (7.81)       232.6841 (5.59)      4,057.8555 (7.85)       139.8660 (4.06)        21;20    242.5261 (0.13)        238           1
test_gzip[fireworks.jpeg-used-output_len=False-gzip]                    3,225.5690 (6.47)      5,195.1360 (5.85)      3,414.4240 (6.47)       234.7346 (5.64)      3,347.1610 (6.47)       129.6175 (3.76)        25;26    292.8752 (0.15)        285           1
test_gzip[fireworks.jpeg-used-output_len=True-cramjam]                  3,937.0940 (7.90)      6,554.8870 (7.38)      4,114.5315 (7.79)       247.2504 (5.94)      4,073.0180 (7.88)       147.3470 (4.27)          7;7    243.0410 (0.13)        131           1
test_gzip[fireworks.jpeg-used-output_len=True-gzip]                     3,198.9300 (6.42)      5,349.8960 (6.03)      3,404.4112 (6.45)       244.9796 (5.89)      3,344.0150 (6.47)       147.7962 (4.29)        19;18    293.7366 (0.16)        279           1
test_gzip[geo.protodata-used-output_len=False-cramjam]                  1,585.4920 (3.18)      2,736.4010 (3.08)      1,661.7925 (3.15)       102.5084 (2.46)      1,640.5520 (3.17)        73.7418 (2.14)        37;31    601.7598 (0.32)        597           1
test_gzip[geo.protodata-used-output_len=False-gzip]                     2,139.4700 (4.29)      3,134.1810 (3.53)      2,243.2067 (4.25)       124.7361 (3.00)      2,215.2610 (4.28)        91.4940 (2.65)        33;21    445.7904 (0.24)        442           1
test_gzip[geo.protodata-used-output_len=True-cramjam]                   1,581.0920 (3.17)      2,745.9970 (3.09)      1,656.5480 (3.14)       113.7994 (2.73)      1,627.7865 (3.15)        66.2845 (1.92)        41;41    603.6650 (0.32)        612           1
test_gzip[geo.protodata-used-output_len=True-gzip]                      2,150.2560 (4.31)      3,237.2160 (3.65)      2,254.2443 (4.27)       149.2253 (3.59)      2,217.4510 (4.29)        92.2995 (2.68)        30;29    443.6076 (0.23)        437           1
test_gzip[html-used-output_len=False-cramjam]                           1,960.5420 (3.93)      3,141.3660 (3.54)      2,048.0674 (3.88)       109.5254 (2.63)      2,023.5200 (3.91)        78.3633 (2.27)        35;27    488.2652 (0.26)        495           1
test_gzip[html-used-output_len=False-gzip]                              2,473.2450 (4.96)      3,558.9080 (4.01)      2,594.7358 (4.91)       128.5785 (3.09)      2,570.0210 (4.97)       103.8637 (3.01)        33;20    385.3957 (0.20)        379           1
test_gzip[html-used-output_len=True-cramjam]                            1,958.1090 (3.93)      3,045.2460 (3.43)      2,048.7684 (3.88)       118.1116 (2.84)      2,024.7970 (3.92)        73.6460 (2.14)        25;25    488.0981 (0.26)        460           1
test_gzip[html-used-output_len=True-gzip]                               2,459.9110 (4.93)      3,534.8000 (3.98)      2,597.0396 (4.92)       153.6156 (3.69)      2,559.4020 (4.95)        97.3550 (2.82)        31;30    385.0538 (0.20)        387           1
test_gzip[html_x_4-used-output_len=False-cramjam]                       9,386.4630 (18.82)    10,598.7610 (11.94)     9,729.7528 (18.42)      273.3211 (6.57)      9,641.9175 (18.64)      232.4350 (6.74)         18;9    102.7775 (0.05)        100           1
test_gzip[html_x_4-used-output_len=False-gzip]                         10,954.5380 (21.97)    12,229.1090 (13.78)    11,388.3416 (21.56)      314.4430 (7.56)     11,291.9835 (21.84)      441.7480 (12.81)        28;0     87.8091 (0.05)         88           1
test_gzip[html_x_4-used-output_len=True-cramjam]                        9,362.4440 (18.78)    10,832.3570 (12.20)     9,753.7416 (18.47)      237.0947 (5.70)      9,740.3650 (18.83)      287.9575 (8.35)         20;4    102.5248 (0.05)         99           1
test_gzip[html_x_4-used-output_len=True-gzip]                          10,925.2580 (21.91)    12,430.4210 (14.00)    11,443.7624 (21.67)      264.1632 (6.35)     11,395.8990 (22.04)      319.3560 (9.26)         22;2     87.3838 (0.05)         87           1
test_gzip[kppkn.gtb-used-output_len=False-cramjam]                     39,132.5940 (78.48)    40,894.0630 (46.07)    39,748.0645 (75.26)      372.1767 (8.94)     39,747.7660 (76.86)      304.5605 (8.83)          6;1     25.1585 (0.01)         25           1
test_gzip[kppkn.gtb-used-output_len=False-gzip]                        63,357.7700 (127.06)   64,867.0860 (73.07)    63,931.2074 (121.05)     431.4790 (10.37)    63,917.2790 (123.60)     455.9250 (13.22)         5;2     15.6418 (0.01)         16           1
test_gzip[kppkn.gtb-used-output_len=True-cramjam]                      39,015.7680 (78.25)    42,225.1180 (47.56)    39,948.5314 (75.64)      763.3876 (18.34)    39,720.7660 (76.81)      784.1145 (22.73)         6;2     25.0322 (0.01)         25           1
test_gzip[kppkn.gtb-used-output_len=True-gzip]                         63,601.0280 (127.55)   65,441.2490 (73.72)    64,262.9742 (121.68)     533.9574 (12.83)    64,180.3945 (124.10)     821.0655 (23.81)         5;0     15.5611 (0.01)         16           1
test_gzip[lcet10.txt-used-output_len=False-cramjam]                    26,893.0920 (53.93)    28,413.5420 (32.01)    27,353.3749 (51.79)      383.5836 (9.22)     27,217.9150 (52.63)      500.9510 (14.52)         9;2     36.5586 (0.02)         37           1
test_gzip[lcet10.txt-used-output_len=False-gzip]                       29,453.3110 (59.07)    32,535.7740 (36.65)    30,217.1804 (57.22)      580.9736 (13.96)    30,086.3775 (58.18)      727.5420 (21.09)         6;1     33.0938 (0.02)         34           1
test_gzip[lcet10.txt-used-output_len=True-cramjam]                     26,531.6900 (53.21)    27,972.4670 (31.51)    27,264.7757 (51.63)      359.6821 (8.64)     27,271.8585 (52.73)      541.0055 (15.69)        12;0     36.6774 (0.02)         36           1
test_gzip[lcet10.txt-used-output_len=True-gzip]                        29,633.5980 (59.43)    31,146.0220 (35.08)    30,229.0893 (57.24)      363.9110 (8.74)     30,206.0570 (58.41)      489.7215 (14.20)        10;0     33.0807 (0.02)         33           1
test_gzip[paper-100k.pdf-used-output_len=False-cramjam]                 2,830.4720 (5.68)      3,538.2900 (3.99)      2,964.3271 (5.61)       107.5574 (2.58)      2,941.1885 (5.69)       120.8570 (3.50)        64;15    337.3447 (0.18)        318           1
test_gzip[paper-100k.pdf-used-output_len=False-gzip]                    3,236.7040 (6.49)      4,065.8930 (4.58)      3,425.2506 (6.49)       142.2416 (3.42)      3,401.9400 (6.58)       147.6630 (4.28)        69;16    291.9494 (0.15)        290           1
test_gzip[paper-100k.pdf-used-output_len=True-cramjam]                  2,829.1540 (5.67)      3,389.4550 (3.82)      2,957.7600 (5.60)       108.1616 (2.60)      2,929.9220 (5.67)       113.1508 (3.28)        58;24    338.0937 (0.18)        331           1
test_gzip[paper-100k.pdf-used-output_len=True-gzip]                     3,244.5330 (6.51)      5,180.7690 (5.84)      3,436.5281 (6.51)       201.4648 (4.84)      3,392.0260 (6.56)       125.4383 (3.64)        26;26    290.9914 (0.15)        261           1
test_gzip[plrabn12.txt-used-output_len=False-cramjam]                  49,320.9690 (98.91)    74,427.5540 (83.84)    53,266.0228 (100.86)   5,882.3811 (141.35)   51,283.4770 (99.17)    2,127.7145 (61.69)         2;3     18.7737 (0.01)         20           1
test_gzip[plrabn12.txt-used-output_len=False-gzip]                     58,243.8110 (116.81)   62,047.1040 (69.89)    60,300.9332 (114.18)   1,419.1243 (34.10)    60,473.5410 (116.94)   2,739.9335 (79.44)         7;0     16.5835 (0.01)         16           1
test_gzip[plrabn12.txt-used-output_len=True-cramjam]                   49,100.9320 (98.47)    63,868.7720 (71.95)    52,104.9511 (98.66)    4,869.3169 (117.00)   49,982.7855 (96.65)    1,824.6450 (52.90)         3;3     19.1920 (0.01)         16           1
test_gzip[plrabn12.txt-used-output_len=True-gzip]                      58,425.0160 (117.17)   63,285.9410 (71.29)    59,995.6096 (113.60)   1,132.8758 (27.22)    60,199.5660 (116.41)   1,314.8300 (38.12)         4;1     16.6679 (0.01)         17           1
test_gzip[urls.10K-used-output_len=False-cramjam]                      36,073.3510 (72.34)    41,615.1860 (46.88)    38,465.0136 (72.83)    1,701.3922 (40.88)    38,056.7860 (73.59)    3,375.4912 (97.87)        10;0     25.9977 (0.01)         23           1
test_gzip[urls.10K-used-output_len=False-gzip]                         40,484.9950 (81.19)    42,920.4170 (48.35)    41,712.5562 (78.98)      718.7713 (17.27)    41,697.8900 (80.63)    1,178.5225 (34.17)         7;0     23.9736 (0.01)         25           1
test_gzip[urls.10K-used-output_len=True-cramjam]                       36,067.1570 (72.33)    39,557.0310 (44.56)    37,054.0519 (70.16)      729.0753 (17.52)    36,985.7040 (71.52)      730.4390 (21.18)         7;1     26.9876 (0.01)         27           1
test_gzip[urls.10K-used-output_len=True-gzip]                          40,734.0010 (81.69)    43,980.7760 (49.54)    42,111.1947 (79.74)      836.1285 (20.09)    42,234.7625 (81.67)    1,114.0880 (32.30)         9;0     23.7467 (0.01)         24           1
-----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
```

#### Snappy

There are two types here, `framed` and `raw`; the recommended one being snappy's `framed` format.

`make bench-snappy-framed`

```bash
------------------------------------------------------------------------------------------------------------------------- benchmark: 28 tests -------------------------------------------------------------------------------------------------------------------------
Name (time in us)                                                    Min                       Max                      Mean                    StdDev                    Median                     IQR            Outliers          OPS            Rounds  Iterations
-----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
test_snappy_framed[Mark.Twain-Tom.Sawyer.txt-cramjam]            66.3110 (1.0)            132.0670 (1.0)             71.5088 (1.0)              8.0543 (1.0)             68.8310 (1.0)            2.5820 (1.0)       472;777  13,984.2887 (1.0)        4909           1
test_snappy_framed[Mark.Twain-Tom.Sawyer.txt-snappy]            128.1280 (1.93)           280.2570 (2.12)           141.0132 (1.97)            13.1797 (1.64)           136.6060 (1.98)          10.3185 (4.00)      445;314   7,091.5369 (0.51)       3901           1
test_snappy_framed[alice29.txt-cramjam]                         667.8120 (10.07)        1,122.6380 (8.50)           728.5935 (10.19)           70.0799 (8.70)           709.3380 (10.31)         42.2965 (16.38)     152;155   1,372.5075 (0.10)       1407           1
test_snappy_framed[alice29.txt-snappy]                        1,363.7070 (20.57)        1,933.3930 (14.64)        1,463.5245 (20.47)           79.0356 (9.81)         1,446.3070 (21.01)         80.3440 (31.12)      101;31     683.2820 (0.05)        544           1
test_snappy_framed[asyoulik.txt-cramjam]                        580.7250 (8.76)         1,038.6950 (7.86)           614.7546 (8.60)            42.1717 (5.24)           604.4420 (8.78)          39.3620 (15.24)      152;95   1,626.6652 (0.12)       1648           1
test_snappy_framed[asyoulik.txt-snappy]                       1,158.6120 (17.47)        1,628.6620 (12.33)        1,227.1376 (17.16)           68.0189 (8.45)         1,213.2820 (17.63)         52.7795 (20.44)       89;68     814.9045 (0.06)        780           1
test_snappy_framed[fifty-four-mb-random-cramjam]             77,210.1300 (>1000.0)     81,672.0980 (618.41)      78,480.3618 (>1000.0)      1,149.2224 (142.68)      78,176.9650 (>1000.0)      885.1638 (342.82)        2;1      12.7420 (0.00)         13           1
test_snappy_framed[fifty-four-mb-random-snappy]           5,559,679.8660 (>1000.0)  8,568,313.0780 (>1000.0)  7,917,293.5574 (>1000.0)  1,319,007.0957 (>1000.0)  8,465,963.1640 (>1000.0)  830,941.5408 (>1000.0)       1;1       0.1263 (0.00)          5           1
test_snappy_framed[fifty-four-mb-repeating-cramjam]          53,625.9030 (808.70)      56,103.9790 (424.81)      54,536.0511 (762.65)         865.3173 (107.44)      54,204.6480 (787.50)     1,152.4509 (446.34)        5;0      18.3365 (0.00)         18           1
test_snappy_framed[fifty-four-mb-repeating-snappy]          408,793.3940 (>1000.0)    413,610.6820 (>1000.0)    411,025.2038 (>1000.0)      2,027.2435 (251.70)     410,251.0240 (>1000.0)    3,348.1738 (>1000.0)       2;0       2.4329 (0.00)          5           1
test_snappy_framed[fireworks.jpeg-cramjam]                       93.1590 (1.40)           229.4800 (1.74)           102.4807 (1.43)            11.1360 (1.38)            99.3520 (1.44)           6.0148 (2.33)      863;919   9,757.9334 (0.70)       8747           1
test_snappy_framed[fireworks.jpeg-snappy]                       634.3260 (9.57)           969.7800 (7.34)           670.6240 (9.38)            41.6387 (5.17)           659.5460 (9.58)          43.4297 (16.82)      158;81   1,491.1486 (0.11)       1377           1
test_snappy_framed[geo.protodata-cramjam]                       202.6670 (3.06)           448.9580 (3.40)           216.7288 (3.03)            19.7819 (2.46)           211.0980 (3.07)          12.7265 (4.93)      409;402   4,614.0605 (0.33)       4327           1
test_snappy_framed[geo.protodata-snappy]                        725.8820 (10.95)        1,141.0820 (8.64)           772.4960 (10.80)           44.3410 (5.51)           768.2020 (11.16)         35.2232 (13.64)      269;96   1,294.5051 (0.09)       1283           1
test_snappy_framed[html-cramjam]                                210.3770 (3.17)           385.4390 (2.92)           223.8841 (3.13)            20.1802 (2.51)           213.4705 (3.10)          14.0190 (5.43)      456;391   4,466.5962 (0.32)       3874           1
test_snappy_framed[html-snappy]                                 661.9320 (9.98)         1,073.4110 (8.13)           699.2566 (9.78)            44.5358 (5.53)           687.3445 (9.99)          46.4350 (17.98)      151;66   1,430.0901 (0.10)       1360           1
test_snappy_framed[html_x_4-cramjam]                            799.6210 (12.06)        1,471.1290 (11.14)          876.7533 (12.26)           94.8879 (11.78)          851.4875 (12.37)         51.6310 (20.00)     118;123   1,140.5717 (0.08)       1202           1
test_snappy_framed[html_x_4-snappy]                           2,718.3700 (40.99)        3,591.7680 (27.20)        2,945.9552 (41.20)          133.4797 (16.57)        2,910.0090 (42.28)        125.9050 (48.76)       77;28     339.4485 (0.02)        334           1
test_snappy_framed[kppkn.gtb-cramjam]                           570.4850 (8.60)           942.4960 (7.14)           604.2190 (8.45)            37.0589 (4.60)           599.5330 (8.71)          39.6085 (15.34)      157;79   1,655.0290 (0.12)       1568           1
test_snappy_framed[kppkn.gtb-snappy]                          1,418.7900 (21.40)        1,968.2600 (14.90)        1,498.6198 (20.96)           70.0844 (8.70)         1,487.2870 (21.61)         58.7050 (22.74)      144;47     667.2807 (0.05)        619           1
test_snappy_framed[lcet10.txt-cramjam]                        1,753.5210 (26.44)        2,618.3510 (19.83)        1,927.7128 (26.96)          161.5183 (20.05)        1,880.7385 (27.32)        111.6060 (43.22)       78;55     518.7495 (0.04)        494           1
test_snappy_framed[lcet10.txt-snappy]                         3,772.7230 (56.89)        4,574.4210 (34.64)        4,001.1636 (55.95)          142.0375 (17.63)        3,985.7400 (57.91)        172.0220 (66.62)        65;9     249.9273 (0.02)        239           1
test_snappy_framed[paper-100k.pdf-cramjam]                       71.3860 (1.08)           282.0590 (2.14)            81.4524 (1.14)            10.4778 (1.30)            79.0920 (1.15)           6.5700 (2.54)      918;899  12,277.1087 (0.88)       9596           1
test_snappy_framed[paper-100k.pdf-snappy]                       533.3570 (8.04)           821.8800 (6.22)           563.8659 (7.89)            35.0754 (4.35)           556.9420 (8.09)          38.9318 (15.08)      146;75   1,773.4712 (0.13)       1609           1
test_snappy_framed[plrabn12.txt-cramjam]                      2,542.3630 (38.34)        3,281.8630 (24.85)        2,737.9968 (38.29)          111.7024 (13.87)        2,720.9985 (39.53)        138.1680 (53.51)       89;11     365.2305 (0.03)        354           1
test_snappy_framed[plrabn12.txt-snappy]                       5,096.0520 (76.85)        6,126.9320 (46.39)        5,441.2705 (76.09)          190.2210 (23.62)        5,427.8250 (78.86)        247.3365 (95.79)        53;3     183.7806 (0.01)        181           1
test_snappy_framed[urls.10K-cramjam]                          2,274.0300 (34.29)        3,110.8600 (23.56)        2,521.7336 (35.26)          130.5882 (16.21)        2,493.3860 (36.22)        139.6460 (54.08)      103;21     396.5526 (0.03)        396           1
test_snappy_framed[urls.10K-snappy]                           5,976.5190 (90.13)        6,779.0760 (51.33)        6,312.5501 (88.28)          177.9607 (22.10)        6,286.4100 (91.33)        233.4860 (90.43)        49;1     158.4146 (0.01)        152           1
-----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
```

`make bench-snappy-raw`

```bash
------------------------------------------------------------------------------------------------------------- benchmark: 28 tests --------------------------------------------------------------------------------------------------------------
Name (time in us)                                              Min                    Max                   Mean                StdDev                 Median                 IQR            Outliers          OPS            Rounds  Iterations
------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
test_snappy_raw[Mark.Twain-Tom.Sawyer.txt-cramjam]         51.3180 (3.37)        106.5710 (2.12)         54.9889 (3.29)         6.7249 (2.09)         52.6220 (3.32)       1.8468 (2.57)      627;721  18,185.5023 (0.30)       5755           1
test_snappy_raw[Mark.Twain-Tom.Sawyer.txt-snappy]          52.5980 (3.45)        110.6400 (2.20)         56.2056 (3.36)         6.7379 (2.10)         53.6730 (3.39)       1.6682 (2.32)    1269;1560  17,791.8193 (0.30)      11281           1
test_snappy_raw[alice29.txt-cramjam]                      599.5640 (39.35)       955.8480 (18.98)       634.3983 (37.97)       39.8069 (12.38)       624.0110 (39.38)     40.4120 (56.29)      134;81   1,576.2968 (0.03)       1469           1
test_snappy_raw[alice29.txt-snappy]                       600.3480 (39.41)     1,046.0080 (20.77)       641.8216 (38.42)       52.8951 (16.45)       626.3260 (39.52)     41.8230 (58.25)     148;121   1,558.0655 (0.03)       1631           1
test_snappy_raw[asyoulik.txt-cramjam]                     531.3770 (34.88)       950.9300 (18.88)       567.7280 (33.98)       53.5173 (16.65)       553.5235 (34.93)     38.8605 (54.13)     150;129   1,761.4068 (0.03)       1664           1
test_snappy_raw[asyoulik.txt-snappy]                      532.8600 (34.98)       826.0870 (16.41)       566.0961 (33.88)       42.7028 (13.28)       553.0790 (34.90)     38.2127 (53.23)       97;70   1,766.4846 (0.03)        981           1
test_snappy_raw[fifty-four-mb-random-cramjam]          64,900.1350 (>1000.0)  66,259.3000 (>1000.0)  65,499.6381 (>1000.0)    488.8944 (152.06)   65,556.7080 (>1000.0)  900.7005 (>1000.0)       9;0      15.2673 (0.00)         16           1
test_snappy_raw[fifty-four-mb-random-snappy]           55,702.4500 (>1000.0)  63,684.2990 (>1000.0)  56,684.4748 (>1000.0)  1,802.6270 (560.68)   56,194.9530 (>1000.0)  807.8321 (>1000.0)       1;1      17.6415 (0.00)         18           1
test_snappy_raw[fifty-four-mb-repeating-cramjam]       61,115.8000 (>1000.0)  63,571.1790 (>1000.0)  62,403.5144 (>1000.0)    668.1779 (207.83)   62,227.2180 (>1000.0)  969.2140 (>1000.0)       7;0      16.0247 (0.00)         17           1
test_snappy_raw[fifty-four-mb-repeating-snappy]        32,122.7550 (>1000.0)  33,752.5920 (670.29)   32,768.2958 (>1000.0)    441.6249 (137.36)   32,726.9150 (>1000.0)  592.4183 (825.17)       12;0      30.5173 (0.00)         31           1
test_snappy_raw[fireworks.jpeg-cramjam]                    24.0780 (1.58)        185.5780 (3.69)         25.9138 (1.55)         4.1937 (1.30)         24.5250 (1.55)       0.8781 (1.22)    2273;2403  38,589.5241 (0.64)      25542           1
test_snappy_raw[fireworks.jpeg-snappy]                     15.2350 (1.0)          50.3550 (1.0)          16.7069 (1.0)          3.2151 (1.0)          15.8470 (1.0)        0.8130 (1.13)    1175;1223  59,855.6622 (1.0)       15115           1
test_snappy_raw[geo.protodata-cramjam]                    155.5390 (10.21)       383.3330 (7.61)        165.6040 (9.91)        16.6777 (5.19)        160.6820 (10.14)      9.3510 (13.02)     576;655   6,038.5013 (0.10)       5400           1
test_snappy_raw[geo.protodata-snappy]                     142.9101 (9.38)        268.0700 (5.32)        152.4274 (9.12)        12.7332 (3.96)        147.1700 (9.29)       8.9320 (12.44)     669;569   6,560.5007 (0.11)       6050           1
test_snappy_raw[html-cramjam]                             165.7090 (10.88)       331.3710 (6.58)        176.3145 (10.55)       16.6429 (5.18)        171.4330 (10.82)     10.0080 (13.94)     573;606   5,671.6825 (0.09)       5450           1
test_snappy_raw[html-snappy]                              156.6590 (10.28)       376.6970 (7.48)        166.9049 (9.99)        16.5424 (5.15)        162.0930 (10.23)      9.4865 (13.21)     576;624   5,991.4352 (0.10)       5383           1
test_snappy_raw[html_x_4-cramjam]                         662.5010 (43.49)     1,178.5590 (23.40)       702.4728 (42.05)       58.7407 (18.27)       685.1415 (43.23)     41.8875 (58.34)      110;94   1,423.5426 (0.02)       1140           1
test_snappy_raw[html_x_4-snappy]                          632.7050 (41.53)     1,083.1810 (21.51)       676.2055 (40.47)       59.6778 (18.56)       658.1900 (41.53)     43.0540 (59.97)     147;126   1,478.8404 (0.02)       1453           1
test_snappy_raw[kppkn.gtb-cramjam]                        500.7480 (32.87)       863.8660 (17.16)       528.4980 (31.63)       37.5560 (11.68)       517.9300 (32.68)     32.8555 (45.76)      143;96   1,892.1548 (0.03)       1692           1
test_snappy_raw[kppkn.gtb-snappy]                         504.1240 (33.09)       855.1690 (16.98)       532.1999 (31.86)       36.5428 (11.37)       522.2600 (32.96)     34.3139 (47.80)     160;103   1,878.9933 (0.03)       1834           1
test_snappy_raw[lcet10.txt-cramjam]                     1,585.7740 (104.09)    2,621.0520 (52.05)     1,683.1401 (100.75)     123.0068 (38.26)     1,654.2825 (104.39)    82.9070 (115.48)      45;40     594.1276 (0.01)        538           1
test_snappy_raw[lcet10.txt-snappy]                      1,590.9740 (104.43)    2,517.1320 (49.99)     1,686.2719 (100.93)     104.7156 (32.57)     1,664.4925 (105.04)    74.3340 (103.54)      58;42     593.0242 (0.01)        582           1
test_snappy_raw[paper-100k.pdf-cramjam]                    27.2180 (1.79)        198.5880 (3.94)         29.5554 (1.77)         4.9856 (1.55)         27.9830 (1.77)       0.9690 (1.35)    2051;2213  33,834.7155 (0.57)      20315           1
test_snappy_raw[paper-100k.pdf-snappy]                     20.1430 (1.32)         57.0260 (1.13)         21.8590 (1.31)         3.2314 (1.01)         21.0150 (1.33)       0.7179 (1.0)     1753;1970  45,747.6927 (0.76)      26477           1
test_snappy_raw[plrabn12.txt-cramjam]                   2,393.2550 (157.09)    3,470.2860 (68.92)     2,615.1149 (156.53)     126.0164 (39.20)     2,584.6890 (163.10)   128.8117 (179.42)      59;17     382.3924 (0.01)        385           1
test_snappy_raw[plrabn12.txt-snappy]                    2,185.9910 (143.48)    3,105.9010 (61.68)     2,312.9505 (138.44)     122.6518 (38.15)     2,280.8130 (143.93)    94.9631 (132.27)      48;32     432.3482 (0.01)        394           1
test_snappy_raw[urls.10K-cramjam]                       1,851.8370 (121.55)    2,722.8100 (54.07)     1,961.6530 (117.42)     100.0348 (31.11)     1,936.3300 (122.19)    84.8327 (118.16)      67;27     509.7741 (0.01)        413           1
test_snappy_raw[urls.10K-snappy]                        1,813.8860 (119.06)    2,805.6660 (55.72)     2,047.6586 (122.56)     121.4122 (37.76)     2,024.1400 (127.73)    93.0140 (129.56)      75;44     488.3627 (0.01)        498           1
------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
```


##### Benchmarks for using `compress_into` and `decompress_into`
All variants except `lz4` for now, implement a `compress_into` and `decompress_into`
This is where, if you have a Python numpy array pre-allocated to the size of the
de/compress data output, we can place the bytes directly into that buffer.

Again, since basically no variants implement similar functionality as we saw in the Gzip `output_len`
benchmarks, this benchmark is specific to `cramjam`

```bash
--------------------------------------------------------------------------------------------------------------------------- benchmark: 28 tests ---------------------------------------------------------------------------------------------------------------------------
Name (time in us)                                                                           Min                    Max                   Mean              StdDev                 Median                 IQR            Outliers          OPS            Rounds  Iterations
---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
test_cramjam_snappy_de_compress_into[Mark.Twain-Tom.Sawyer.txt-compress_into]           39.5430 (1.89)        226.3750 (4.12)         42.8299 (1.93)       6.0547 (2.06)         41.0640 (1.93)       1.4650 (1.96)    2225;2454  23,348.1766 (0.52)      18846           1
test_cramjam_snappy_de_compress_into[Mark.Twain-Tom.Sawyer.txt-decompress_into]         20.8690 (1.0)          54.9380 (1.0)          22.1814 (1.0)        2.9375 (1.0)          21.2590 (1.0)        0.7462 (1.0)       463;481  45,082.7264 (1.0)        7705           1
test_cramjam_snappy_de_compress_into[alice29.txt-compress_into]                        463.9240 (22.23)       777.6850 (14.16)       492.9410 (22.22)     32.7011 (11.13)       483.5000 (22.74)     31.2150 (41.83)     197;115   2,028.6405 (0.04)       2053           1
test_cramjam_snappy_de_compress_into[alice29.txt-decompress_into]                      186.3790 (8.93)        382.7770 (6.97)        199.2822 (8.98)      21.8619 (7.44)        192.5040 (9.06)      11.0680 (14.83)     436;575   5,018.0092 (0.11)       4754           1
test_cramjam_snappy_de_compress_into[asyoulik.txt-compress_into]                       412.3180 (19.76)       713.2160 (12.98)       438.1621 (19.75)     36.8647 (12.55)       426.4940 (20.06)     29.6507 (39.73)     226;170   2,282.2606 (0.05)       2321           1
test_cramjam_snappy_de_compress_into[asyoulik.txt-decompress_into]                     163.3530 (7.83)        314.8180 (5.73)        173.4995 (7.82)      18.4431 (6.28)        165.1880 (7.77)       9.7200 (13.03)     567;637   5,763.7058 (0.13)       5530           1
test_cramjam_snappy_de_compress_into[fifty-four-mb-random-compress_into]            21,370.5300 (>1000.0)  23,421.1310 (426.32)   22,271.0823 (>1000.0)  581.2127 (197.86)   22,076.6660 (>1000.0)  839.1198 (>1000.0)       9;0      44.9013 (0.00)         25           1
test_cramjam_snappy_de_compress_into[fifty-four-mb-random-decompress_into]          14,683.7930 (703.62)   15,945.4020 (290.24)   15,056.4776 (678.79)   266.1213 (90.59)    15,000.5550 (705.61)   379.6751 (508.79)        7;1      66.4166 (0.00)         30           1
test_cramjam_snappy_de_compress_into[fifty-four-mb-repeating-compress_into]         14,899.9090 (713.97)   16,099.5890 (293.05)   15,429.4449 (695.60)   276.9102 (94.27)    15,329.7200 (721.09)   336.6940 (451.19)       20;0      64.8111 (0.00)         63           1
test_cramjam_snappy_de_compress_into[fifty-four-mb-repeating-decompress_into]       17,915.9630 (858.50)   19,089.1940 (347.47)   18,389.6426 (829.06)   278.4235 (94.78)    18,373.7230 (864.28)   418.8805 (561.32)       11;0      54.3784 (0.00)         29           1
test_cramjam_snappy_de_compress_into[fireworks.jpeg-compress_into]                      43.9390 (2.11)        118.4420 (2.16)         47.8558 (2.16)       5.6904 (1.94)         46.2460 (2.18)       1.9707 (2.64)    1996;2197  20,896.1152 (0.46)      19715           1
test_cramjam_snappy_de_compress_into[fireworks.jpeg-decompress_into]                    29.7500 (1.43)         73.8140 (1.34)         31.8771 (1.44)       3.9980 (1.36)         30.5820 (1.44)       1.2860 (1.72)    2432;2732  31,370.4800 (0.70)      27124           1
test_cramjam_snappy_de_compress_into[geo.protodata-compress_into]                      121.8950 (5.84)        260.3290 (4.74)        129.7199 (5.85)      12.8547 (4.38)        124.3525 (5.85)       7.0450 (9.44)      799;848   7,708.9151 (0.17)       7196           1
test_cramjam_snappy_de_compress_into[geo.protodata-decompress_into]                     61.7630 (2.96)        152.9090 (2.78)         65.6442 (2.96)       7.3924 (2.52)         62.6280 (2.95)       2.0600 (2.76)    1269;1729  15,233.6341 (0.34)      12907           1
test_cramjam_snappy_de_compress_into[html-compress_into]                               129.4560 (6.20)        252.4930 (4.60)        138.6755 (6.25)      14.4065 (4.90)        133.9925 (6.30)       7.5980 (10.18)    834;1040   7,211.0811 (0.16)       7318           1
test_cramjam_snappy_de_compress_into[html-decompress_into]                              62.3400 (2.99)        154.2940 (2.81)         66.4681 (3.00)       8.0530 (2.74)         64.3110 (3.03)       3.5030 (4.69)    1064;1480  15,044.8144 (0.33)      11675           1
test_cramjam_snappy_de_compress_into[html_x_4-compress_into]                           517.2050 (24.78)       888.9890 (16.18)       570.9151 (25.74)     50.9777 (17.35)       554.5600 (26.09)     44.1403 (59.15)     353;194   1,751.5740 (0.04)       1857           1
test_cramjam_snappy_de_compress_into[html_x_4-decompress_into]                         257.5690 (12.34)       529.3130 (9.63)        283.4639 (12.78)     37.2057 (12.67)       266.7790 (12.55)     26.8391 (35.97)     301;283   3,527.7862 (0.08)       2446           1
test_cramjam_snappy_de_compress_into[kppkn.gtb-compress_into]                          363.2850 (17.41)       611.8500 (11.14)       383.6018 (17.29)     26.0204 (8.86)        375.6370 (17.67)     23.2070 (31.10)     295;182   2,606.8697 (0.06)       2684           1
test_cramjam_snappy_de_compress_into[kppkn.gtb-decompress_into]                        195.4920 (9.37)        389.6650 (7.09)        206.6503 (9.32)      18.2140 (6.20)        201.5670 (9.48)      11.5785 (15.52)     518;513   4,839.0933 (0.11)       4671           1
test_cramjam_snappy_de_compress_into[lcet10.txt-compress_into]                       1,230.0970 (58.94)     1,987.3830 (36.18)     1,324.3744 (59.71)     91.6116 (31.19)     1,300.6140 (61.18)     70.7862 (94.86)       91;53     755.0735 (0.02)        671           1
test_cramjam_snappy_de_compress_into[lcet10.txt-decompress_into]                       489.8850 (23.47)       866.4870 (15.77)       531.4879 (23.96)     58.0469 (19.76)       516.8210 (24.31)     32.6948 (43.81)     126;136   1,881.5104 (0.04)       1385           1
test_cramjam_snappy_de_compress_into[paper-100k.pdf-compress_into]                      38.2830 (1.83)         97.7700 (1.78)         42.4763 (1.91)       5.6442 (1.92)         40.7560 (1.92)       2.0365 (2.73)    2347;2564  23,542.5644 (0.52)      19856           1
test_cramjam_snappy_de_compress_into[paper-100k.pdf-decompress_into]                    29.9950 (1.44)         80.4800 (1.46)         32.0370 (1.44)       4.0740 (1.39)         31.0030 (1.46)       0.9880 (1.32)    2199;2610  31,213.8860 (0.69)      26692           1
test_cramjam_snappy_de_compress_into[plrabn12.txt-compress_into]                     1,670.0190 (80.02)     2,474.3870 (45.04)     1,797.8845 (81.05)    104.9018 (35.71)     1,773.9745 (83.45)     89.2560 (119.61)     104;39     556.2093 (0.01)        580           1
test_cramjam_snappy_de_compress_into[plrabn12.txt-decompress_into]                     657.2370 (31.49)     1,143.2750 (20.81)       709.0987 (31.97)     61.9860 (21.10)       697.8510 (32.83)     33.1955 (44.48)      87;104   1,410.2409 (0.03)       1088           1
test_cramjam_snappy_de_compress_into[urls.10K-compress_into]                         1,454.5130 (69.70)     2,186.5280 (39.80)     1,609.2278 (72.55)    129.4296 (44.06)     1,569.5480 (73.83)    100.0898 (134.13)      54;36     621.4161 (0.01)        467           1
test_cramjam_snappy_de_compress_into[urls.10K-decompress_into]                         612.5320 (29.35)     1,063.2690 (19.35)       670.1230 (30.21)     62.6793 (21.34)       654.0500 (30.77)     43.8632 (58.78)      104;89   1,492.2634 (0.03)       1135           1
---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
```

#### Lz4

`make bench-lz4`

```bash
---------------------------------------------------------------------------------------------------------- benchmark: 24 tests ----------------------------------------------------------------------------------------------------------
Name (time in us)                                          Min                    Max                   Mean              StdDev                 Median                 IQR            Outliers         OPS            Rounds  Iterations
-----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
test_lz4[Mark.Twain-Tom.Sawyer.txt-cramjam]           265.4710 (1.29)      1,709.2960 (3.48)        283.5229 (1.30)      36.6529 (1.45)        273.4540 (1.29)      20.1355 (1.62)      288;307  3,527.0517 (0.77)       3672           1
test_lz4[Mark.Twain-Tom.Sawyer.txt-python-lz4]        206.1390 (1.0)         491.0620 (1.0)         218.6584 (1.0)       25.2966 (1.0)         212.6340 (1.0)       12.4433 (1.0)         71;88  4,573.3440 (1.0)         969           1
test_lz4[alice29.txt-cramjam]                       1,591.6270 (7.72)      2,422.1640 (4.93)      1,670.5685 (7.64)     105.3038 (4.16)      1,641.9690 (7.72)      69.4932 (5.58)        47;41    598.5986 (0.13)        583           1
test_lz4[alice29.txt-python-lz4]                    3,014.7940 (14.63)     5,390.9890 (10.98)     3,217.0799 (14.71)    294.7618 (11.65)     3,150.8370 (14.82)    147.8563 (11.88)       20;26    310.8409 (0.07)        315           1
test_lz4[asyoulik.txt-cramjam]                      1,341.9520 (6.51)      2,609.5330 (5.31)      1,435.5756 (6.57)     147.8273 (5.84)      1,403.8710 (6.60)      73.3415 (5.89)        39;52    696.5847 (0.15)        700           1
test_lz4[asyoulik.txt-python-lz4]                   2,624.6560 (12.73)     4,103.2770 (8.36)      2,832.4582 (12.95)    198.1849 (7.83)      2,785.6950 (13.10)    156.1410 (12.55)       42;25    353.0502 (0.08)        351           1
test_lz4[fireworks.jpeg-cramjam]                      279.7980 (1.36)      1,311.0030 (2.67)        312.3522 (1.43)      38.4700 (1.52)        300.8090 (1.41)      26.9838 (2.17)      231;178  3,201.5137 (0.70)       2419           1
test_lz4[fireworks.jpeg-python-lz4]                 2,608.3740 (12.65)     4,068.8360 (8.29)      2,757.1025 (12.61)    187.2245 (7.40)      2,709.8345 (12.74)     98.5295 (7.92)        26;28    362.6996 (0.08)        340           1
test_lz4[geo.protodata-cramjam]                       540.2170 (2.62)        954.5990 (1.94)        576.6946 (2.64)      59.2219 (2.34)        558.5890 (2.63)      40.7170 (3.27)      135;113  1,734.0201 (0.38)       1662           1
test_lz4[geo.protodata-python-lz4]                    670.0600 (3.25)      1,260.5120 (2.57)        706.9607 (3.23)      64.6132 (2.55)        691.1930 (3.25)      34.7968 (2.80)        62;67  1,414.5058 (0.31)       1371           1
test_lz4[html-cramjam]                                557.7070 (2.71)        952.2150 (1.94)        589.8092 (2.70)      40.2866 (1.59)        580.9055 (2.73)      41.6970 (3.35)       131;55  1,695.4636 (0.37)       1478           1
test_lz4[html-python-lz4]                             693.6250 (3.36)      1,243.8680 (2.53)        731.9760 (3.35)      57.6403 (2.28)        717.0350 (3.37)      42.7220 (3.43)        79;68  1,366.1650 (0.30)       1381           1
test_lz4[html_x_4-cramjam]                          1,834.1350 (8.90)      3,089.3180 (6.29)      1,947.9892 (8.91)     130.8457 (5.17)      1,913.9850 (9.00)      93.6955 (7.53)        41;30    513.3499 (0.11)        480           1
test_lz4[html_x_4-python-lz4]                       2,873.1570 (13.94)     4,251.4200 (8.66)      3,030.2486 (13.86)    164.9341 (6.52)      2,984.9600 (14.04)    127.3245 (10.23)       25;20    330.0059 (0.07)        301           1
test_lz4[kppkn.gtb-cramjam]                         1,491.9130 (7.24)      2,781.5860 (5.66)      1,570.2841 (7.18)      87.5343 (3.46)      1,553.1860 (7.30)      72.3630 (5.82)        49;32    636.8274 (0.14)        571           1
test_lz4[kppkn.gtb-python-lz4]                      3,199.8520 (15.52)     4,217.5750 (8.59)      3,359.6494 (15.36)    124.7223 (4.93)      3,325.6840 (15.64)    118.4477 (9.52)        39;21    297.6501 (0.07)        297           1
test_lz4[lcet10.txt-cramjam]                        4,124.9650 (20.01)     5,752.2620 (11.71)     4,313.8181 (19.73)    200.0237 (7.91)      4,261.9530 (20.04)    164.7980 (13.24)        16;9    231.8132 (0.05)        149           1
test_lz4[lcet10.txt-python-lz4]                     8,179.7110 (39.68)     9,336.4010 (19.01)     8,561.2170 (39.15)    252.1781 (9.97)      8,493.2420 (39.94)    365.4150 (29.37)        27;1    116.8058 (0.03)        115           1
test_lz4[paper-100k.pdf-cramjam]                      430.1570 (2.09)        892.4910 (1.82)        460.2170 (2.10)      50.9308 (2.01)        445.1600 (2.09)      32.3620 (2.60)      109;104  2,172.8878 (0.48)       1670           1
test_lz4[paper-100k.pdf-python-lz4]                 1,705.9020 (8.28)      3,126.1760 (6.37)      1,840.7896 (8.42)     179.6683 (7.10)      1,791.6345 (8.43)      88.1440 (7.08)        53;65    543.2451 (0.12)        570           1
test_lz4[plrabn12.txt-cramjam]                      5,020.0420 (24.35)     6,864.6530 (13.98)     5,292.4698 (24.20)    253.9889 (10.04)     5,205.3250 (24.48)    209.6257 (16.85)       18;16    188.9477 (0.04)        161           1
test_lz4[plrabn12.txt-python-lz4]                  11,864.6600 (57.56)    13,558.0060 (27.61)    12,295.0743 (56.23)    312.6359 (12.36)    12,212.9405 (57.44)    337.8520 (27.15)        19;4     81.3334 (0.02)         78           1
test_lz4[urls.10K-cramjam]                          4,636.7460 (22.49)     5,603.8480 (11.41)     4,867.3892 (22.26)    182.9029 (7.23)      4,814.5570 (22.64)    200.3785 (16.10)        27;6    205.4489 (0.04)        125           1
test_lz4[urls.10K-python-lz4]                       9,835.2580 (47.71)    12,131.1310 (24.70)    10,280.7427 (47.02)    417.8376 (16.52)    10,100.7300 (47.50)    393.7080 (31.64)        14;6     97.2692 (0.02)         74           1
-----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
```

#### Brotli

`make bench-brotli`

```bash
----------------------------------------------------------------------------------------------------- benchmark: 24 tests ------------------------------------------------------------------------------------------------------
Name (time in ms)                                         Min                   Max                  Mean             StdDev                Median                IQR            Outliers      OPS            Rounds  Iterations
--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
test_brotli[Mark.Twain-Tom.Sawyer.txt-brotli]         19.3134 (1.05)        27.2349 (1.30)        21.3094 (1.10)      2.6975 (4.93)        19.9170 (1.03)      2.5291 (4.19)          4;2  46.9276 (0.91)         19           1
test_brotli[Mark.Twain-Tom.Sawyer.txt-cramjam]        18.4340 (1.0)         21.0191 (1.0)         19.3708 (1.0)       0.5474 (1.0)         19.3012 (1.0)       0.6043 (1.0)          19;1  51.6241 (1.0)          49           1
test_brotli[alice29.txt-brotli]                      224.1868 (12.16)      227.8789 (10.84)      225.7909 (11.66)     1.3771 (2.52)       225.6526 (11.69)     1.7266 (2.86)          2;0   4.4289 (0.09)          5           1
test_brotli[alice29.txt-cramjam]                     231.9518 (12.58)      234.0236 (11.13)      233.0551 (12.03)     0.9388 (1.72)       233.5518 (12.10)     1.6091 (2.66)          2;0   4.2908 (0.08)          5           1
test_brotli[asyoulik.txt-brotli]                     179.9550 (9.76)       184.0013 (8.75)       182.1853 (9.41)      1.4951 (2.73)       182.4221 (9.45)      2.1882 (3.62)          2;0   5.4889 (0.11)          6           1
test_brotli[asyoulik.txt-cramjam]                    184.3326 (10.00)      186.8581 (8.89)       185.3278 (9.57)      0.9119 (1.67)       185.2473 (9.60)      1.0866 (1.80)          2;0   5.3958 (0.10)          6           1
test_brotli[fireworks.jpeg-brotli]                    71.5389 (3.88)        75.3505 (3.58)        73.0897 (3.77)      0.9971 (1.82)        72.9035 (3.78)      0.9791 (1.62)          4;1  13.6818 (0.27)         14           1
test_brotli[fireworks.jpeg-cramjam]                   72.5284 (3.93)        79.9106 (3.80)        73.9391 (3.82)      1.8016 (3.29)        73.4397 (3.80)      0.7829 (1.30)          1;1  13.5246 (0.26)         14           1
test_brotli[geo.protodata-brotli]                    126.7453 (6.88)       131.5479 (6.26)       127.8292 (6.60)      1.5408 (2.81)       127.3805 (6.60)      0.6052 (1.00)          1;1   7.8229 (0.15)          8           1
test_brotli[geo.protodata-cramjam]                   130.8282 (7.10)       133.5629 (6.35)       131.9438 (6.81)      0.9135 (1.67)       131.6701 (6.82)      1.2637 (2.09)          3;0   7.5790 (0.15)          8           1
test_brotli[html-brotli]                             134.3089 (7.29)       136.8154 (6.51)       135.5654 (7.00)      0.9767 (1.78)       135.6640 (7.03)      1.7840 (2.95)          4;0   7.3765 (0.14)          8           1
test_brotli[html-cramjam]                            137.2950 (7.45)       141.9890 (6.76)       138.5814 (7.15)      1.6047 (2.93)       137.9258 (7.15)      1.2205 (2.02)          1;1   7.2160 (0.14)          7           1
test_brotli[html_x_4-brotli]                         164.2704 (8.91)       167.2311 (7.96)       165.5355 (8.55)      1.1562 (2.11)       165.5598 (8.58)      2.0072 (3.32)          3;0   6.0410 (0.12)          6           1
test_brotli[html_x_4-cramjam]                        167.8992 (9.11)       172.8526 (8.22)       169.7993 (8.77)      1.6606 (3.03)       169.4014 (8.78)      0.9248 (1.53)          2;1   5.8893 (0.11)          6           1
test_brotli[kppkn.gtb-brotli]                        427.2876 (23.18)      436.4801 (20.77)      431.6189 (22.28)     3.8362 (7.01)       430.7206 (22.32)     6.4873 (10.73)         2;0   2.3169 (0.04)          5           1
test_brotli[kppkn.gtb-cramjam]                       466.2677 (25.29)      497.4289 (23.67)      478.3085 (24.69)    13.1943 (24.10)      475.5884 (24.64)    21.5743 (35.70)         1;0   2.0907 (0.04)          5           1
test_brotli[lcet10.txt-brotli]                       745.3311 (40.43)      848.7560 (40.38)      799.6870 (41.28)    48.8617 (89.26)      809.2787 (41.93)    93.6578 (154.98)        2;0   1.2505 (0.02)          5           1
test_brotli[lcet10.txt-cramjam]                      762.7996 (41.38)      880.4882 (41.89)      798.9608 (41.25)    47.1400 (86.11)      788.5498 (40.85)    44.9762 (74.43)         1;0   1.2516 (0.02)          5           1
test_brotli[paper-100k.pdf-brotli]                   404.2915 (21.93)      409.0611 (19.46)      406.8449 (21.00)     1.9765 (3.61)       407.1359 (21.09)     3.3423 (5.53)          2;0   2.4579 (0.05)          5           1
test_brotli[paper-100k.pdf-cramjam]                  406.3386 (22.04)      410.8830 (19.55)      408.1782 (21.07)     1.8307 (3.34)       407.8163 (21.13)     2.7956 (4.63)          2;0   2.4499 (0.05)          5           1
test_brotli[plrabn12.txt-brotli]                     784.4900 (42.56)      797.3552 (37.93)      789.0102 (40.73)     4.9220 (8.99)       787.9514 (40.82)     4.6725 (7.73)          1;0   1.2674 (0.02)          5           1
test_brotli[plrabn12.txt-cramjam]                    793.1241 (43.02)      845.8696 (40.24)      816.3642 (42.14)    26.2341 (47.92)      802.2609 (41.57)    48.5658 (80.37)         2;0   1.2249 (0.02)          5           1
test_brotli[urls.10K-brotli]                       1,251.5425 (67.89)    1,265.7612 (60.22)    1,257.9783 (64.94)     6.7086 (12.26)    1,256.7268 (65.11)    12.8291 (21.23)         1;0   0.7949 (0.02)          5           1
test_brotli[urls.10K-cramjam]                      1,290.7796 (70.02)    1,365.9914 (64.99)    1,322.5113 (68.27)    33.0967 (60.46)    1,304.2875 (67.58)    54.4744 (90.14)         1;0   0.7561 (0.01)          5           1
--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
```

#### ZSTD

`make bench-zstd`

```bash 
------------------------------------------------------------------------------------------------------- benchmark: 24 tests -------------------------------------------------------------------------------------------------------
Name (time in us)                                       Min                   Max                  Mean              StdDev                Median                 IQR            Outliers         OPS            Rounds  Iterations
-----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
test_zstd[Mark.Twain-Tom.Sawyer.txt-cramjam]       154.4260 (1.0)        429.8220 (1.61)       165.7084 (1.0)       18.0132 (1.0)        159.4600 (1.0)       11.8590 (1.0)       211;207  6,034.6953 (1.0)        1685           1
test_zstd[Mark.Twain-Tom.Sawyer.txt-zstd]          191.9860 (1.24)       267.6270 (1.0)        219.1654 (1.32)      28.9777 (1.61)       197.9910 (1.24)      56.6957 (4.78)         22;0  4,562.7633 (0.76)         53           1
test_zstd[alice29.txt-cramjam]                   1,362.2300 (8.82)     2,102.8580 (7.86)     1,439.4675 (8.69)      73.2922 (4.07)     1,422.7575 (8.92)      70.1800 (5.92)       134;36    694.7014 (0.12)        684           1
test_zstd[alice29.txt-zstd]                      1,315.1700 (8.52)     2,365.9580 (8.84)     1,403.9396 (8.47)     120.7466 (6.70)     1,372.8835 (8.61)      70.7280 (5.96)        72;76    712.2813 (0.12)        744           1
test_zstd[asyoulik.txt-cramjam]                  1,211.2850 (7.84)     1,841.2200 (6.88)     1,284.1431 (7.75)      77.5367 (4.30)     1,270.1210 (7.97)      59.2710 (5.00)        78;53    778.7294 (0.13)        757           1
test_zstd[asyoulik.txt-zstd]                     1,157.6700 (7.50)     1,653.1200 (6.18)     1,220.0159 (7.36)      65.1603 (3.62)     1,204.0080 (7.55)      55.9547 (4.72)        92;49    819.6615 (0.14)        805           1
test_zstd[fireworks.jpeg-cramjam]                  292.3470 (1.89)       602.9980 (2.25)       313.0014 (1.89)      33.5971 (1.87)       302.4510 (1.90)      20.7312 (1.75)      287;295  3,194.8734 (0.53)       3057           1
test_zstd[fireworks.jpeg-zstd]                     216.5420 (1.40)       452.4620 (1.69)       229.1745 (1.38)      22.7349 (1.26)       222.9190 (1.40)      12.9210 (1.09)      378;420  4,363.4868 (0.72)       3623           1
test_zstd[geo.protodata-cramjam]                   324.0560 (2.10)       672.1640 (2.51)       354.0909 (2.14)      42.7156 (2.37)       340.2680 (2.13)      27.8100 (2.35)      216;209  2,824.1338 (0.47)       2621           1
test_zstd[geo.protodata-zstd]                      281.7450 (1.82)       535.1640 (2.00)       300.6179 (1.81)      29.3579 (1.63)       291.3200 (1.83)      18.6490 (1.57)      247;242  3,326.4816 (0.55)       2894           1
test_zstd[html-cramjam]                            349.8810 (2.27)       769.5230 (2.88)       376.6673 (2.27)      43.1311 (2.39)       362.2320 (2.27)      27.6060 (2.33)      188;186  2,654.8631 (0.44)       2332           1
test_zstd[html-zstd]                               312.4120 (2.02)       580.8570 (2.17)       333.4397 (2.01)      32.0637 (1.78)       323.5910 (2.03)      21.6778 (1.83)      253;222  2,999.0433 (0.50)       2851           1
test_zstd[html_x_4-cramjam]                        476.8870 (3.09)       887.5280 (3.32)       537.3586 (3.24)      76.9432 (4.27)       509.1320 (3.19)      46.7175 (3.94)      115;122  1,860.9548 (0.31)        909           1
test_zstd[html_x_4-zstd]                           364.0010 (2.36)       797.9640 (2.98)       389.8506 (2.35)      37.9180 (2.11)       377.5145 (2.37)      28.5240 (2.41)      215;177  2,565.0855 (0.43)       2332           1
test_zstd[kppkn.gtb-cramjam]                     1,082.0570 (7.01)     1,696.9090 (6.34)     1,148.6073 (6.93)      72.4650 (4.02)     1,131.7090 (7.10)      57.8217 (4.88)        85;62    870.6196 (0.14)        749           1
test_zstd[kppkn.gtb-zstd]                        1,043.9460 (6.76)     1,771.5120 (6.62)     1,110.4534 (6.70)      81.5983 (4.53)     1,090.0220 (6.84)      54.9060 (4.63)        95;82    900.5331 (0.15)        879           1
test_zstd[lcet10.txt-cramjam]                    3,509.2670 (22.72)    5,409.0240 (20.21)    3,734.3465 (22.54)    230.2715 (12.78)    3,684.9145 (23.11)    163.8595 (13.82)       19;14    267.7845 (0.04)        268           1
test_zstd[lcet10.txt-zstd]                       3,387.3740 (21.94)    5,608.0490 (20.95)    3,595.1312 (21.70)    246.2549 (13.67)    3,542.5960 (22.22)    180.9382 (15.26)       19;12    278.1540 (0.05)        283           1
test_zstd[paper-100k.pdf-cramjam]                  488.0500 (3.16)       942.0480 (3.52)       520.3597 (3.14)      44.4527 (2.47)       508.0725 (3.19)      36.1535 (3.05)      166;116  1,921.7476 (0.32)       1900           1
test_zstd[paper-100k.pdf-zstd]                     479.8510 (3.11)       868.9010 (3.25)       520.1514 (3.14)      46.2761 (2.57)       509.5520 (3.20)      28.6118 (2.41)      197;190  1,922.5171 (0.32)       1977           1
test_zstd[plrabn12.txt-cramjam]                  4,742.2590 (30.71)    6,472.1400 (24.18)    5,019.8933 (30.29)    234.2066 (13.00)    4,958.0030 (31.09)    198.3953 (16.73)       26;12    199.2074 (0.03)        167           1
test_zstd[plrabn12.txt-zstd]                     4,586.0060 (29.70)    6,362.5020 (23.77)    4,846.4730 (29.25)    238.6297 (13.25)    4,803.3360 (30.12)    206.9840 (17.45)        26;9    206.3356 (0.03)        203           1
test_zstd[urls.10K-cramjam]                      4,455.0140 (28.85)    8,134.9550 (30.40)    4,870.4095 (29.39)    491.9910 (27.31)    4,727.1960 (29.65)    212.0393 (17.88)       23;28    205.3215 (0.03)        201           1
test_zstd[urls.10K-zstd]                         4,177.0250 (27.05)    5,455.3670 (20.38)    4,437.3242 (26.78)    210.9385 (11.71)    4,382.1610 (27.48)    231.2377 (19.50)       44;11    225.3610 (0.04)        227           1
-----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
```




